/*
 * Copyright (C) 2010 Julien VAUBOURG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

package tortues;

import java.awt.Point;
import java.util.*;
import java.io.*;
import javax.swing.JFileChooser;
import java.net.URISyntaxException;

/** Controller (MVC)
  * @author <a href="http://julien.vaubourg.com">Julien VAUBOURG</a>
  * @version 0.2 (2010-10-12)
  */
public class Controller {
	private ArrayList<Point[]> segments;
	private ArrayList<String> history;
	private int historyLevel, iScenario, orientationReplay;
	private Tortoise franklin;
	private Screen view;
	private Point pointReplay;

	// PROTECTED

	/** Creates the controller, with a turtle
	  * @param franklin Tortoise instance of Franklin
	  */
	protected Controller(Tortoise franklin) {
		this.franklin = franklin;
		segments = new ArrayList<Point[]>();
		initHistory();
		pointReplay = (Point) franklin.getCoord().clone();
		orientationReplay = franklin.getOrientation();
		iScenario = 0;
	}

	/** Returns the drawing
	  * @return Segments composing the drawing
	  */
	protected ArrayList<Point[]> getSegments() {
		return segments;
	}

	/** Returns Franklin
	  * @return Franklin tortoise instance
	  */
	protected Tortoise getFranklin() {
		return franklin;
	}

	/** Back in the command history
	  * @return Older command
	  */
	protected String historyUp() {
		return !history.isEmpty() && historyLevel > 0 ?
			history.get(--historyLevel) :
			"";
	}

	/** Goes down in the command history
	  * @return Newer command
	  */
	protected String historyDown() {
		if(historyLevel == history.size() - 1)
			historyLevel++;

		return historyLevel < history.size() ?
			history.get(++historyLevel) :
			"";
	}

	/** Set the attached view (MVC)
	  * @param view Screen view instance
	  */
	protected void setView(Screen view) {
		this.view = view;
	}

	/** Executes a turtle command, and adds it to the history
	  * @param commandText Turtle command
	  */
	protected void execCommand(String commandText) {
		execCommand(commandText, true);
	}

	/** Executes a turtle command
	  * @param commandText Turtle command
	  * @param save Save in the command history
	  */
	protected void execCommand(String commandText, boolean save) {
		String[] commandTextDetails = commandText.split("\\s+");
		boolean hasArgument = false;
		int commandArgument = 0, distanceLeft;
		Point oldCoord;

		// Throw an IllegalArgumentException if commandText is not a
		// known command
		Commands command = Commands.valueOf(commandTextDetails[0]);

		// Add to command history
		if(save && command != Commands.REPLAY
			&& command != Commands.SAVE
			&& command != Commands.CLEAR)
			addHistory(commandText);

		// Command with argument
		if(commandTextDetails.length > 1) {
			hasArgument = true;

			// Throw a NumberFormatException if the argument is not
			// a number
			commandArgument =
				Integer.parseInt(commandTextDetails[1]);
		}

		// Commands
		switch(command) {

			// Forward and Backward
			case FD :
			case BD :

				// Argument required
				if(!hasArgument && commandArgument > 0)
					throw new IllegalArgumentException();

				// While it's left a distance to draw (it's a
				// torique plan)
				do {

					// Save current position
					oldCoord = (Point) franklin.getCoord().clone();

					// Forward
					if(command == Commands.FD)

						// Forward until that Franklin
						// strikes a border
						distanceLeft = franklin.forward(
							commandArgument,
							view.getVivariumSize()
						);
					else
						// Backward until that Franklin
						// strikes a border
						distanceLeft = franklin.backward(
							commandArgument,
							view.getVivariumSize()
						);

					// If the pen is down, the traveled path
					// is saved as a segment
					if(franklin.isInDrawMode()) {
						Point[] nl = {
							oldCoord,
							(Point) franklin.getCoord().clone()
						};

						segments.add(nl);
					}

					// Franklin has striked a border, it
					// must to continue, thanks to the
					// torique plan
					if(distanceLeft > 0) {
						try {
							// Put Franklin on the
							// corresponding point,
							// on the torique plan
							if(command == Commands.FD)
								franklin.forwardTorique(
									view.getVivariumSize()
								);
							else
								franklin.backwardTorique(
									view.getVivariumSize()
								);

							// With the do-while,
							// Franklin can to
							// continue its way
							commandArgument = distanceLeft;

						// If the torique plan is
						// abnormal
						} catch(IllegalArgumentException e) {
							System.out.println(e.getMessage());
							System.exit(1);
						}
					}

				} while(distanceLeft > 0);
			break;

			// Turn right
			case RT :
				// Argument required
				if(!hasArgument || commandArgument < 0)
					throw new IllegalArgumentException();

				franklin.turnRight(commandArgument);
			break;

			// Turn left
			case LT :
				// Argument required
				if(!hasArgument || commandArgument < 0)
					throw new IllegalArgumentException();

				franklin.turnLeft(commandArgument);
			break;

			// Pen up (drawing mode disabled)
			case PENUP :
				franklin.penUp();
			break;

			// Pen down (drawing mode enabled)
			case PENDOWN :
				franklin.penDown();
			break;

			// Clear drawing (erase segments and the command
			// history)
			case CLEAR :
				segments.clear();
				initHistory();
				pointReplay = (Point) franklin.getCoord().clone();
				orientationReplay = franklin.getOrientation();
			break;

			// Change pen color
			case BC :
				// Argument required
				if(!hasArgument || commandArgument < 0 || commandArgument > 4)
					throw new IllegalArgumentException();

				view.setPenColor(commandArgument);
			break;

			// Oh! What is it?
			case ANSWER :
				view.answer();
			break;

			// Exit software
			case EXIT :
				System.exit(0);
			break;

			// Reset Franklin
			case ORIGIN :
				franklin.origin();
			break;

			// Put Franklin to the center
			case CENTER :
				franklin.go2(new Point(
					view.getSize().width / 2,
					view.getSize().height / 2
				));
			break;

			// Replay the current drawing
			case REPLAY :
				replay();
			break;

			// Save the drawing
			case SAVE :
				save();
			break;

			// Open a saved drawing
			case OPEN :
				open();
			break;

			// Display help (on console)
			case HELP :
				Controller.help();
			break;
		}
	}

	/** Open a saved drawing (with file chooser) */
	protected void open() {
		File dumpFile;

		JFileChooser fc = new JFileChooser();
		fc.showOpenDialog(view);

		dumpFile = fc.getSelectedFile();
		open(dumpFile);
	}

	/** Open a saved drawing
	  * @param dumpFile Saved drawing file
	  */
	protected void open(File dumpFile) {
		FileInputStream dumpStream;

		try {
			// An InputStream is required for JAR files
			dumpStream = new FileInputStream(dumpFile);
			open(dumpStream);

		} catch(FileNotFoundException e) {
			view.wrongCommand("File not found");
		}
	}

	/** Open a saved drawing
	  * @param dumpStream Saved drawing file
	  */
	protected void open(InputStream dumpStream) {
		Scanner dump;
		String[] initReplay;

		// If the user has cliked on Cancel, with the file
		// chooser, File is wrong.
		if(dumpStream instanceof InputStream) {

			// Read dump
			dump = new Scanner(dumpStream);

			// Clear history
			initHistory();

			// Where et how this scenario begins
			initReplay = dump.nextLine().split(",");

			pointReplay = new Point(
				Integer.parseInt(initReplay[0]),
				Integer.parseInt(initReplay[1])
			);

			orientationReplay = Integer.parseInt(initReplay[2]);

			// Inject all saved commands
			while(dump.hasNextLine())
				history.add(dump.nextLine());

			dump.close();

			// Play the drawing
			replay();
		}
	}

	/** Help on console */
	static protected void help() {

			// Read
			Scanner man = new Scanner(
				Controller.class.getResourceAsStream("HELP")
			);

			// Print on console
			while(man.hasNextLine())
				System.out.println(man.nextLine());

			man.close();
	}

	// PRIVATE

	/** Saves a command to the command history
	  * @param command Turtle command
	  */
	private void addHistory(String command) {
		history.add(command);
		historyLevel = history.size();
	}

	/** Init the command history */
	private void initHistory() {
		history = new ArrayList<String>();
		historyLevel = 0;
	}

	/** Is view (MVC) defined, with setView
	  * @return Is view defined
	  */
	private boolean isViewDefined() {
		return view instanceof Screen;
	}

	/** Replay the drawing (executes all commands in the history) */
	private void replay() {
		Timer scenario = new Timer();

		// Its a recursive function : on init, the screen is reset
		if(iScenario == 0) {
			franklin.go2(pointReplay);
			execCommand("PENUP", false);
			execCommand("RT " +
				(360 - franklin.getOrientation() + orientationReplay)
			);
			segments.clear();
		}

		// Task
		TimerTask play = new TimerTask() {
			public void run() {

				// If there are commands left
				if(iScenario < history.size()) {

					// Execute a command
					execCommand(history.get(iScenario++), false);
					view.repaintVivarium();

					// Recursion
					replay();

				// Replaying done
				} else
					iScenario = 0;
			}
		};

		// The task is executed every 30ms
		GregorianCalendar stopScenario = new GregorianCalendar();
		stopScenario.add(stopScenario.MILLISECOND, 30);
		scenario.schedule(play, stopScenario.getTime());
	}

	/** Save the current drawing */
	private void save() {
		File dumpFile;
		FileWriter dump;

		// Graphical file chooser
		JFileChooser fc = new JFileChooser();
		fc.showSaveDialog(view);

		try {
			dumpFile = fc.getSelectedFile();

			// If the user has not clicked on Cancel
			if(dumpFile instanceof File) {

				// Writer
				dump = new FileWriter(dumpFile);

				// Where and how this scenario begins
				dump.write(pointReplay.x + ","
					+ pointReplay.y + ","
					+ orientationReplay + "\n");

				// Writing of all commands in the file
				for(String c : history)
					dump.write(c + "\n");

				dump.close();
			}

		} catch(FileNotFoundException e) {
			view.wrongCommand("File not found or insufficient permissions");

		} catch(IOException e) {
			view.wrongCommand("Can not write this file");
		}
	}
}
