/*
 * Copyright (C) 2010 Julien VAUBOURG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

package tortues;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.util.Timer;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import javax.imageio.ImageIO;
import java.io.File;
import java.net.URISyntaxException;

/** View (MVC)
  * @author <a href="http://julien.vaubourg.com">Julien VAUBOURG</a>
  * @version 0.2 (2010-10-12)
  */
public class Screen extends JFrame {
	private JPanel world, vivarium;
	private JTextField commandLine;
	private BufferedImage franklinFace;
	private int curPenColor;
	private final Color[] penColors = {
		Color.WHITE,
		Color.YELLOW,
		Color.BLUE,
		Color.RED,
		Color.GREEN
	};

	private Controller controller;
	private Tortoise franklin;
	private int answerCommand;

	// PROTECTED

	/** Creates the view (MVC), with the controller
	  * @param size Desired window size
	  * @param c Controller
	  */
	protected Screen(Dimension size, Controller c) {
		this.controller = c;
		franklin = controller.getFranklin();

		setTitle("Franklin");
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

		// Main window
		world = new JPanel();
		world.setLayout(new BorderLayout());

		// The Franklin's world
		vivarium = createVivarium();
		commandLine = new JTextField();

		// The command line (textfield)
		commandLine.addKeyListener(
			new KeyListener() {

				// The textfield is like a shell, you can browse
				// in your old commands with ARROW and DOWN
				// ARROW
				public void keyPressed(KeyEvent e) {
					String command;
					int k = e.getKeyCode();

					// ARROW : lastest command
					if(k == e.VK_UP) {
						command = controller.historyUp();

						if(!command.isEmpty())
							commandLine.setText(command);

					// DOWN ARROW : newer command
					} else if(k == e.VK_DOWN) {
						command = controller.historyDown();
						commandLine.setText(command);

					// ESCAPE : focus the vivarium
					} else if(k == e.VK_ESCAPE) {
						vivarium.setFocusable(true);
						vivarium.requestFocus();
					}
				}

				public void keyTyped(KeyEvent e) {}
				public void keyReleased(KeyEvent e) {}
			}
		);

		// When ENTER is pressed on the command line
		commandLine.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				String commandText = commandLine.getText();

				try {

					execCommand(commandText);

				// This command is not in the command list
				} catch(NumberFormatException e) {

					wrongCommand("Wrong command");

				// This argument is not a number (or a wrong
				// number)
				} catch(IllegalArgumentException e) {

					wrongCommand("Wrong argument");

				}
			}
		});

		world.add(vivarium, BorderLayout.CENTER);
		world.add(commandLine, BorderLayout.SOUTH);
		getContentPane().add(world);

		setVisible(true);
		setSize(size.width, size.height);

		// Centering on the screen
		setLocation(
			Math.max(0, (screenSize.width  - getSize().width) / 2),
			Math.max(0, (screenSize.height - getSize().height) / 2)
		);

		// Default pen color is blue
		curPenColor = 2;

		setDefaultCloseOperation(EXIT_ON_CLOSE);
	}

	/** Returns the vivarium size
	  * @return Vivarium size
	  */
	protected Dimension getVivariumSize() {
		return vivarium.getSize();
	}

	/** Set the pen color (with index)
	  * @param c Color index
	  */
	protected void setPenColor(int c) {
		curPenColor = c;
	}

	/** Oh! What is it? */
	protected void answer() {
		javax.swing.Timer blink;

		answerCommand = 0;

		commandLine.setFont(commandLine.getFont().deriveFont(Font.BOLD));
		commandLine.setHorizontalAlignment(JTextField.CENTER);

		controller.open(
			// .tts = TurTle Scenario
			getClass().getResourceAsStream("answer.tts")
		);

		// No comment... suprise ^^
		ActionListener blinking = new ActionListener() {
			public void actionPerformed(ActionEvent ae) {
				commandLine.setText(
					(answerCommand % 6 == 0) ?
					"                                                   " :
					"* THE ANSWER TO LIFE, THE UNIVERSE AND EVERYTHING *"
				);

				curPenColor = answerCommand++ % penColors.length;
				vivarium.repaint();
			}
		};

		blink = new javax.swing.Timer(100, blinking);
		blink.start();
	}

	/** Refresh the vivarium */
	protected void repaintVivarium() {
		vivarium.repaint();
	}

	/** When the user typed a wrong command (or a good command with a wrong
	  * argument), he is alerted directly on the command line
	  * @param error Alert message
	  */
	protected void wrongCommand(String error) {

		// The alert message is typed directly in the command line, red over
		// white, bolded and centered
		commandLine.setBackground(Color.RED);
		commandLine.setForeground(Color.WHITE);
		commandLine.setHorizontalAlignment(JTextField.CENTER);
		commandLine.setFont(commandLine.getFont().deriveFont(Font.BOLD));
		commandLine.setText(" * " + error.toUpperCase() + " ! *");

		Timer blink = new Timer();

		// After Xms, the command line returns to a normal state
		TimerTask blinking = new TimerTask() {
			public void run() {
				// Black over white, not bolded, to left
				commandLine.setBackground(Color.WHITE);
				commandLine.setFont(commandLine.getFont().deriveFont(Font.PLAIN));
				commandLine.setForeground(Color.BLACK);
				commandLine.setHorizontalAlignment(JTextField.LEFT);
				commandLine.setText("");
			}
		};

		// Start stopwatch, to clear the message
		GregorianCalendar stopBlink = new GregorianCalendar();
		stopBlink.add(stopBlink.MILLISECOND, 700);
		blink.schedule(blinking, stopBlink.getTime());
	}

	// PRIVATE

	/** Creates the Franklin's world, with all keyboard shortcuts
	  * @return Franklin's world
	  */
	private JPanel createVivarium() {
		JPanel jp;

		try {
			// Get the Franklin's face
			franklinFace = ImageIO.read(
				getClass().getResourceAsStream("img/turtle.png")
			);

		} catch(Exception e) {
			System.err.println("/!\\ Turtle img not found");
			System.exit(1);
		}

		// Franklin's world
		jp = new JPanel() {

			// This function will be called each time that the world
			// will be refreshed
			public void paint(Graphics g) {
				super.paint(g);

				// Rotate the Franklin's face, on the good
				// direction
				BufferedImage franklinFaceTmp = rotateImg(
					franklinFace,
					franklin.getOrientation()
				);

				// Foreground color is the current pen color
				g.setColor(penColors[curPenColor]);

				// Put the Franklin's face where is Franklin
				g.drawImage(
					(Image) franklinFaceTmp,
					franklin.getCoord().x - 12,
					franklin.getCoord().y - 12,
					this
				);

				// If the pen is down, Franklin has a color
				// point on his body
				if(franklin.isInDrawMode())
					g.fillOval(
						franklin.getCoord().x - 4,
						franklin.getCoord().y - 4,
						8, 8
					);

				// Draw all segments of the drawing
				for(Point[] c : controller.getSegments()) {
					((Graphics2D) g).draw(new Line2D.Double(
						c[0].x, c[0].y,
						c[1].x, c[1].y
					));
				}
			}
		};

		// Keyboard shortcuts
		jp.addKeyListener(
			new KeyListener() {
				public void keyPressed(KeyEvent e) {
					int k = e.getKeyCode();

					// ARROW or J = Forward
					if(k == e.VK_J || k == e.VK_UP) {
						execCommand("FD 10"); }

					// DOWN ARROW or K = Backwrad
					else if(k == e.VK_K || k == e.VK_DOWN)
						execCommand("BD 10");

					// LEFT ARROW or H = Turn left
					else if(k == e.VK_H || k == e.VK_LEFT)
						execCommand("LT 10");

					// RIGHT ARROW or L = Turn right
					else if(k == e.VK_L || k == e.VK_RIGHT)
						execCommand("RT 10");

					// SPACE = Pen down/up
					else if(k == e.VK_SPACE)
						if(franklin.isInDrawMode())
							execCommand("PENUP");
						else
							execCommand("PENDOWN");

					// BACKSPACE = Clear vivarium
					else if(k == e.VK_BACK_SPACE)
						execCommand("CLEAR");

					// HOME = Put Franklin to his original
					// position
					else if(k == e.VK_HOME)
						execCommand("ORIGIN");

					// ESCAPE = Exit software
					else if(k == e.VK_ESCAPE)
						execCommand("EXIT");

					// R = Replay the current drawing
					else if(k == e.VK_R)
						execCommand("REPLAY");

					// S = Save the current drawing in a
					// file
					else if(k == e.VK_S)
						execCommand("SAVE");

					// O = Replay a drawing saved
					else if(k == e.VK_O)
						execCommand("OPEN");

					// M = Display help on console
					else if(k == e.VK_M)
						execCommand("HELP");

					// C = Change the pen color
					else if(k == e.VK_C) {
						execCommand(
							"BC " +
							((curPenColor + 1) % penColors.length)
						);
					}
				}

				public void keyTyped(KeyEvent e) {}
				public void keyReleased(KeyEvent e) {}
			}
		);

		// The user can click on the vivarium to focus it
		jp.addMouseListener(
			new MouseAdapter() {
				public void mouseClicked(MouseEvent e) {
					vivarium.setFocusable(true);
					vivarium.requestFocus();
				}
			}
		);

		jp.setBackground(Color.BLACK);
		jp.setFocusable(true);
		jp.requestFocus();

		return jp;
	}

	/** Executes turtle commands, and refresh the vivarium
	  * @param commandText Turtle command
	  */
	private void execCommand(String commandText) {
		controller.execCommand(commandText);

		commandLine.setText("");
		vivarium.repaint();
	}

	/** Rotate an image, with degrees
	  * @param img Image to rotate
	  * @param angle Number of degrees to rotate
	  * @return Rotated image
	  * @see http://www.javalobby.org/articles/ultimate-image/#10
	  */
	private BufferedImage rotateImg(BufferedImage img, int angle) {
		int w = img.getWidth();
		int h = img.getHeight();

		BufferedImage dimg = dimg = new BufferedImage(w, h, img.getType());
		Graphics2D g = dimg.createGraphics();

		g.rotate(Math.toRadians(angle), w/2, h/2);
		g.drawImage(img, null, 0, 0);

		return dimg;
	}
}
