/*
 * Copyright (C) 2010 Julien VAUBOURG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

package tortues;

import java.awt.Point;
import java.awt.Dimension;
import javax.swing.JPanel;

/** Franklin is a Tortoise
  * @author <a href="http://julien.vaubourg.com">Julien VAUBOURG</a>
  * @version 0.2 (2010-10-12)
  */
public class Tortoise {
	private Point coord;
	private int orientation;
	private boolean penState;
	
	// PROTECTED

	/** Constructor */
	protected Tortoise() {
		origin();
		penState = false;
	}

	/** Put on the original position */
	protected void origin() {
		orientation = 135;
		coord = new Point(42, 42);
	}

	/** Returns the current turtle position
	  * @return Turtle position
	  */
	protected Point getCoord() {
		return coord;
	}

	/** Returns the current turtle orientation
	  * @return Turtle orientation
	  */
	protected int getOrientation() {
		return orientation;
	}

	/** Is pen down ?
	  * @return Pen state
	  */
	protected boolean isInDrawMode() {
		return this.penState;
	}

	/** Change the orientation of the turtle of -x degrees
	  * @param deg Degrees to rotate
	  */
	protected void turnLeft(int deg) {
		turn(-deg);
	}

	/** Change the orientation of the turtle of x degrees
	  * @param deg Degrees to rotate
	  */
	protected void turnRight(int deg) {
		turn(deg);
	}

	/** Switch draw mode to On */
	protected void penDown() {
		penState = true;
	}

	/** Switch draw mode to Off */
	protected void penUp() {
		penState = false;
	}

	/** Forward the turtle according to its orientation
	  * @param steps Number of pixel to move
	  * @param vivariumSize Vivarium's size
	  * @return Pixels left to forward (positive if the turtle has striked a
	  * border) */
	protected int forward(int steps, Dimension vivariumSize) {
		return move(
			steps,
			orientation,
			vivariumSize
		);
	}

	/** Backward the turtle according to its orientation
	  * @param steps Number of pixel to move
	  * @param vivariumSize Vivarium's size
	  * @return Pixels left to backward (positive if the turtle has striked
	  * a border) */
	protected int backward(int steps, Dimension vivariumSize) {
		return move(
			steps,
			// Revert the orientation value
			(orientation + 180) % 360,
			vivariumSize
		);
	}

	/** Put the turtle to a specified position
	  * @param p Go to p
	  */
	protected void go2(Point p) {
		coord = p;
	}

	/** If the turtle is on a border and that it must forward, this function
	  * can to cross it across the torique plan
	  * @param vivariumSize Vivarium's size
	  */
	protected void forwardTorique(Dimension vivariumSize) {
		moveTorique(
			orientation,
			vivariumSize
		);
	}

	/** If the turtle is on a border and that it must backward, this function
	  * can to cross it across the torique plan
	  * @param vivariumSize Vivarium's size
	  */
	protected void backwardTorique(Dimension vivariumSize) {
		moveTorique(
			// Revert the orientation value
			(orientation + 180) % 360,
			vivariumSize
		);
	}

	// PRIVATE

	/** Rotate the turtle orientation, to right
	  * @param turn Number of degrees to rotate
	  */
	private void turn(int deg) {
		int newOrientation = (orientation + deg + 360) % 360;
		orientation = newOrientation;
	}

	/** Is there a border on the turtle way ? If yes, what is the maximum
	  * distance to traverse, before the border ?
	  * @param direction Turtle's orientation
	  * @param vivariumSize Vivarium's size
	  * @return The distance before a border
	  */
	private int distanceBefBeingOut(int direction, Dimension vivariumSize)  {
		double a, b, theta;
		Point pTop, pRight, pBot, pLeft, p1, p2;
		int d1, d2, max;

		switch(direction) {
			// NORTH
			case 0 :
				max = coord.y;
			break;

			// EAST
			case 90 :
				max = vivariumSize.width - coord.x;
			break;

			// SOUTH
			case 180 :
				max = vivariumSize.height - coord.y;
			break;

			// WEST
			case 270 :
				max = coord.x;
			break;

			// Between two
			default :
				// Northwest quarter
				if(direction > 270)
					theta = direction - 270;

				// Northwest quarter
				else if(direction > 180)
					theta = 270 - direction;

				// Southeast quarter
				else if(direction > 90)
					theta = direction - 90;

				// Northeast quarter
				else
					theta = 90 - direction;

				// Slope (fr: coefficient directeur) of the
				// turtle way, according to its orientation
				a = Math.tan(Math.toRadians(theta));

				// Intercept (fr: ordonnee a l'origine)
				b = coord.y - a * coord.x;

				// Intersection point with the top border
				pTop = new Point(
					(int) Math.round(-b / a),
					0
				);

				// Intersection point with the right border
				pRight = new Point(
					vivariumSize.width,
					(int) Math.round(a * vivariumSize.width + b)
				);

				// Intersection point with the bottom border
				pBot = new Point(
					(int) Math.round((vivariumSize.height - b) / a),
					vivariumSize.height
				);

				// Intersection point with the left border
				pLeft = new Point(
					0,
					(int) Math.round(b)
				);

				// The turtle going to the northwest quarter :
				// it can to strike the top or left border
				if(direction > 270) {
					p1 = pTop;
					p2 = pLeft;

				// The turtle going to the southwest quarter :
				// it can to strike the bottom or left border
				} else if(direction > 180) {
					p1 = pBot;
					p2 = pLeft;

				// The turtle going to the southeast quarter :
				// it can to strike the bottom or right border
				} else if(direction > 90) {
					p1 = pBot;
					p2 = pRight;

				// The turtle going to the northeast quarter :
				// it can to strike the top or right border
				} else {
					p1 = pTop;
					p2 = pRight;
				}

				// What is the border that it can strike
				// first ?
				d1 = (int) Math.round(coord.distance(p1));
				d2 = (int) Math.round(coord.distance(p2));

				max = d1 < d2 ? d1 : d2;
		}

		return max;
	}

	/** Moves the turtle, according to its orientation
	  * @param distance Distance to forward
	  * @param direction Turtle's orientation
	  * @param vivariumSize Vivarium's size
	  */
	private int move(int distance, int direction, Dimension vivariumSize) {
		Dimension houseSize;
		int h, distanceMax, distanceLeft, theta;

		// What is the max distance before a border ?
		distanceMax = distanceBefBeingOut(
			direction,
			vivariumSize
		);

		// If we want to going further than the max distance, the
		// distance is shortened and the distance left variable is
		// completed
		if(distance > distanceMax) {
			distanceLeft = distance - distanceMax;
			distance = distanceMax;
		} else
			distanceLeft = 0;

		switch(direction) {
			// NORTH
			case 0 :
				coord.y -= distance;
			break;

			// EAST
			case 90 :
				coord.x += distance;
			break;

			// SOUTH
			case 180 :
				coord.y += distance;
			break;

			// WEST
			case 270 :
				coord.x -= distance;
			break;

			// Between two
			default :

				// Hypothenuse
				h = distance;

				// Theta angle (trigonometry)
				theta = direction;

				// Northwest quarter
				if(theta > 270) {
					theta -= 270;

					// Opposite side
					coord.y -= Math.abs(Math.sin(
						Math.toRadians(theta)))* h;

					// Adjacent side
					coord.x -= Math.abs(Math.cos(
						Math.toRadians(theta))) * h;

				// Southwest quarter
				} else if(theta > 180) {
					theta -= 180;

					// Opposite side
					coord.x -= Math.abs(Math.sin(
						Math.toRadians(theta))) * h;

					// Adjacent side
					coord.y += Math.abs(Math.cos(
						Math.toRadians(theta))) * h;

				// Southeast quarter
				} else if(theta > 90) {
					theta -= 90;

					// Opposite side
					coord.y += Math.abs(Math.sin(
						Math.toRadians(theta))) * h;

					// Adjacent side
					coord.x += Math.abs(Math.cos(
						Math.toRadians(theta))) * h;

				// Northeast side
				} else {

					// Opposite side
					coord.x += Math.abs(Math.sin(
						Math.toRadians(theta))) * h;

					// Adjacent side
					coord.y -= Math.abs(Math.cos(
						Math.toRadians(theta))) * h;
				}
		}

		// Distance left to forward, after a crossing of the torique plan
		return distanceLeft;
	}

	/** If the turtle is on a border and that it must forward, this function
	  * can to cross it across the torique plan
	  * @param direction Turtle's orientation
	  * @param vivariumSize Vivarium's size
	  */
	private void moveTorique(int direction, Dimension vivariumSize)  {

		// The turtle is near the top border and it want to going to north
		if(coord.y <= 2 && (direction > 270 || direction < 90))
			coord.y = vivariumSize.height;

		// The turtle is near the right border and it want to going to east
		else if(Math.abs(vivariumSize.width - coord.x) <= 2 && direction < 180)
			coord.x = 0;

		// The turtle is near the bottom border and it want to going to south
		else if(Math.abs(vivariumSize.height - coord.y) <= 2 && direction < 270 && direction > 90)
			coord.y = 0;
 
		// The turtle is near the left border and it want to going to west
		else if(coord.x <= 2 && direction > 180)
			coord.x = vivariumSize.width;

		// The turtle is not in a torique situation (this can cause an
		// infinite loop)
		else
			throw new IllegalArgumentException(
				"FATAL : Undefined torique situation !");
	}
}
